"""
Provides Nose and Django test case assert functions
"""
from contextlib import contextmanager

from nose.tools import (assert_equal, assert_true, assert_false, assert_in, assert_not_in, assert_raises,
                        assert_not_equal, assert_is, assert_is_instance, assert_greater, assert_less,
                        assert_almost_equal, assert_not_almost_equal, assert_greater_equal, assert_less_equal,
                        assert_not_is_instance, assert_list_equal, assert_tuple_equal, assert_set_equal,
                        assert_dict_equal, assert_sequence_equal, assert_multi_line_equal, assert_is_none,
                        assert_is_not_none, assert_equals, assert_logs)

from germanium import config


if config.TURN_OFF_MAX_DIFF:
    assert_equal.__self__.maxDiff = None


def fail(msg=None):
    raise AssertionError(msg)


@contextmanager
def assert_not_raises(exc_type, func=None, *args, **kwargs):
    try:
        if func:
            func(*args, **kwargs)
        yield None
    except exc_type:
        raise fail('{} raised'.format(exc_type.__name__))



import collections
from functools import wraps

from inspect import isclass, isfunction, ismethod, getfullargspec

import responses
from django.db import transaction
from django.db.models import Model
from django.db.models.fields import DateField, DateTimeField
from django.utils.functional import cached_property


def is_iterable(data):
    return isinstance(data, collections.Iterable) and not isinstance(data, str)


def refresh_model_object(obj):
    obj.refresh_from_db()
    for key, value in obj.__class__.__dict__.items():
        if isinstance(value, cached_property):
            obj.__dict__.pop(key, None)


def refresh_model_objects(*data):
    [refresh_model_object(obj) for obj in data if isinstance(obj, Model) and obj.pk]


def login(function=None, users_generator='get_user', **users_kwargs):
    """Login decorator usage: @login(user_data)"""

    def _login(function):
        def _decorator(self, *args, **kwargs):
            if isinstance(users_generator, str):
                users = getattr(self, users_generator)(**users_kwargs)
            else:
                users = users_generator(self, **users_kwargs)

            if not isinstance(users, (list, tuple)):
                users = (users,)

            for user in users:
                self.login(user)
                function(self, *args, **kwargs)
                self.logout()
        return wraps(function)(_decorator)

    if function:
        return _login(function)
    return _login


def login_all(cls=None, **user_kwargs):
    """Login decorator for all methods inside test usage: @login(user_data)"""

    def _login_all(cls):
        for attr, val in cls.__dict__.items():
            if callable(val) and attr.startswith("test_"):
                setattr(cls, attr, login(**user_kwargs)(val))
        return cls

    if cls:
        return _login_all(cls)
    return _login_all


def data_provider(callable_or_property_or_str, *data_provider_args, **data_provider_kwargs):
    """Data provider decorator, allows another callable to provide the data for the test"""

    def test_decorator(fn):

        def get_data(self, *args):
            callable_or_property = (
                getattr(self, callable_or_property_or_str)
                if isinstance(callable_or_property_or_str, str) else callable_or_property_or_str
            )
            if (isfunction(callable_or_property) and next(iter(getfullargspec(callable_or_property).args),
                                                          None) == 'self'):
                return callable_or_property(self, *data_provider_args, *args, **data_provider_kwargs)
            elif ismethod(callable_or_property) or isclass(callable_or_property) or isfunction(callable_or_property):
                return callable_or_property(*data_provider_args, *args, **data_provider_kwargs)
            else:
                return [list(args) + list(val if is_iterable(val) else [val]) for val in callable_or_property]

        def repl(self, *args):
            data = get_data(self, *args)
            if not is_iterable(data):
                data = (data,)
            for fn_args in data:
                if not is_iterable(fn_args):
                    fn_args = [fn_args]
                if getattr(fn, 'data_provider', False):
                    fn(self, *fn_args)
                else:
                    sid = transaction.savepoint()
                    try:
                        fn(self, *fn_args)
                    except AssertionError:
                        raise
                    finally:
                        transaction.savepoint_rollback(sid)
                        responses.reset()
                        refresh_model_objects(*fn_args)
            repl.data_provider = True
        return wraps(fn)(repl)
    return test_decorator


def turn_off_auto_now(model_class, field_name):

    def _turn_off_auto_now(function):
        def _decorator(self, *args, **kwargs):
                field = model_class._meta.get_field(field_name)
                if not isinstance(field, (DateField, DateTimeField)):
                    raise RuntimeError('Field %s must be DateField or DateTimeField type') % field_name
                if not field.auto_now:
                    raise RuntimeError('Field %s must have set auto_no to True') % field_name
                field.auto_now = False
                try:
                    function(self, *args, **kwargs)
                except AssertionError:
                    raise
                finally:
                    field.auto_now = True
        return wraps(function)(_decorator)

    return _turn_off_auto_now



from io import StringIO

from django.core.management import call_command


def test_call_command(command, *args,  **kwargs):
    call_command(command, stdout=StringIO(), stderr=StringIO(), *args, **kwargs)



import types

from django.conf import settings
from django.test.testcases import TestCase, SimpleTestCase


class GermaniumSimpleTestCaseMixin:

    @classmethod
    def setUpClass(cls):
        super(GermaniumSimpleTestCaseMixin, cls).setUpClass()
        cls.set_up_class()

    @classmethod
    def set_up_class(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        super(GermaniumSimpleTestCaseMixin, cls).tearDownClass()
        cls.tear_down_class()

    @classmethod
    def tear_down_class(cls):
        pass

    def setUp(self):
        super(GermaniumSimpleTestCaseMixin, self).setUp()
        self.set_up()

    def set_up(self):
        pass

    def tearDown(self):
        super(GermaniumSimpleTestCaseMixin, self).tearDown()
        self.tear_down()

    def tear_down(self):
        pass


class GermaniumTestCaseMixin(GermaniumSimpleTestCaseMixin):

    if getattr(settings, 'GERMANIUM_FIXTURES', None):
        fixtures = getattr(settings, 'GERMANIUM_FIXTURES', None)


class GermaniumTestCase(GermaniumTestCaseMixin, TestCase):
    pass


class GermaniumSimpleTestCase(GermaniumSimpleTestCaseMixin, SimpleTestCase):
    pass
